import javax.swing.*;
import java.awt.event.*;
import java.awt.*;
import java.io.*;
import java.util.*;

class Wall extends Point
{
    Wall (int y, int x){
        super (x, y);  // NB: coords inverted
    }
}

class Res extends Point
{
    int id, load;
    Res (int i, int y, int x, int l){
        super (x, y);  // NB: coords inverted
        id = i;
        load = l;
    }
}

class Agent extends Point
{
    int id, load;
    Agent (int i, int y, int x, int l){
        super (x, y);  // NB: coords inverted
        id = i;
        load = l;
    }
}

class ReadFiles
{
    Vector walls, res, agents;
    int envy, envx;
    private int nextCycle = 0;
    private BufferedReader agFile, resFile;
    private String lastLineReadAg, lastLineReadRes;
    boolean foundEOF = false;

    ReadFiles (){
        walls = new Vector ();
        res = new Vector ();
        agents = new Vector ();
        try{
            // initialise env
            readEnv ();
            // open agent file
            agFile = new BufferedReader (new FileReader (Main.agFileName));
            agFile.readLine ();  // unused, env dim
            // open res file
            resFile = new BufferedReader (new FileReader (Main.resFileName));
            resFile.readLine ();  // unused, env dim
        }catch (IOException e){
            e.printStackTrace ();
            System.exit (1);
        }catch (NumberFormatException e){
            e.printStackTrace ();
            System.exit (1);
        }
    }

    protected void finalize (){
        // close agent & res files
        try{
            agFile.close ();
            resFile.close ();
        }catch (IOException e){
            e.printStackTrace ();
            System.exit (1);
        }
    }

    // read env info
    private void readEnv () throws IOException{
        BufferedReader br = new BufferedReader
            (new FileReader (Main.envFileName));
        // read env dim
        String s = br.readLine ();
        StringTokenizer st = new StringTokenizer (s);
        envy = Integer.parseInt (st.nextToken ());
        envx = Integer.parseInt (st.nextToken ());
        // read env
        while ((s = br.readLine ()) != null){
            st = new StringTokenizer (s);
            st.nextToken ();  // unused, "W"
            walls.add (new Wall (Integer.parseInt (st.nextToken ()),
                                 Integer.parseInt (st.nextToken ())));
        }
        br.close ();
    }

    private void addAgent (String s){
        StringTokenizer st = new StringTokenizer (s);
        st.nextToken ();  // unused, "A"
        agents.add (new Agent (Integer.parseInt (st.nextToken ()),  // id
                               Integer.parseInt (st.nextToken ()),  // y
                               Integer.parseInt (st.nextToken ()),  // x
                               Integer.parseInt (st.nextToken ())));  // load
    }

    private void addRes (String s){
        StringTokenizer st = new StringTokenizer (s);
        st.nextToken ();  // unused, "R"
        res.add (new Res (Integer.parseInt (st.nextToken ()),  // id
                          Integer.parseInt (st.nextToken ()),  // y
                          Integer.parseInt (st.nextToken ()),  // x
                          Integer.parseInt (st.nextToken ())));  // load
    }

    private int getCycleTokenAgRes (String s){
        StringTokenizer st = new StringTokenizer (s);
        st.nextToken ();  // unused, "A"
        st.nextToken ();  // unused, agent's id
        st.nextToken ();  // unused, agent's y
        st.nextToken ();  // unused, agent's x
        st.nextToken ();  // unused, agent's load
        return Integer.parseInt (st.nextToken ());  // cycle
    }

    private void resetAgentResFile (){
        // BufferedReader does not allow to reset a stream, hence we reopen it
        try{
            agFile.close ();
            resFile.close ();
            agFile = new BufferedReader (new FileReader (Main.agFileName));
            resFile = new BufferedReader (new FileReader (Main.resFileName));
            agFile.readLine ();  // unused, env dim
            resFile.readLine ();  // unused, env dim
        }catch (IOException e){
            e.printStackTrace ();
            System.exit (1);
        }
    }

    // read agent and res info on the given cycle and fill agents/res vector
    // lateral effects: lastLineReadAg/Res, foundEOF
    void readAgentRes (int cycle){
        try{
            if (cycle < nextCycle)  // need to go backward
                resetAgentResFile ();  // reset, afterwards go forward
            if (cycle != nextCycle){  // need to go forward
                do
                    lastLineReadAg = agFile.readLine ();
                while (getCycleTokenAgRes (lastLineReadAg) < cycle);
                do
                    lastLineReadRes = resFile.readLine ();
                while (getCycleTokenAgRes (lastLineReadRes) < cycle);
            }
            // we are well positioned in the files
            nextCycle = cycle + 1;
            agents.clear ();
            res.clear ();
            boolean feofag, feofres;
            // fill last line read, if empty
            if (lastLineReadAg == null)
                lastLineReadAg = agFile.readLine ();
            // process file
            while (true){
                if (lastLineReadAg == null){  // EOF
                    feofag = true;
                    break;
                }
                if (getCycleTokenAgRes (lastLineReadAg) != cycle){  // end of cycle
                    feofag = false;
                    break;
                }
                addAgent (lastLineReadAg);
                lastLineReadAg = agFile.readLine ();
            }
            // fill last line read, if empty
            if (lastLineReadRes == null)
                lastLineReadRes = resFile.readLine ();
            // process file
            while (true){
                if (lastLineReadRes == null){  // EOF
                    feofres = true;
                    break;
                }
                if (getCycleTokenAgRes (lastLineReadRes) != cycle){  // end of cycle
                    feofres = false;
                    break;
                }
                addRes (lastLineReadRes);
                lastLineReadRes = resFile.readLine ();
            }
            foundEOF = feofag && feofres;
        }catch (IOException e){
            e.printStackTrace ();
            System.exit (1);
        }catch (NumberFormatException e){
            e.printStackTrace ();
            System.exit (2);
        }
    }
}

class MainPanel extends JPanel
{
    ReadFiles files = new ReadFiles ();

    MainPanel (){
        super ();
        files.readAgentRes (Main.startCycle);
        setToolTipText ("Uninit");  // force use of getToolTipText()
    }

    public Dimension getPreferredSize (){
        return new Dimension (files.envx * Main.zoom, files.envy * Main.zoom);
    }

    private void drawWall (Graphics g, Wall w){
        g.setColor (Color.black);
        g.fillRect (w.x * Main.zoom, w.y * Main.zoom, Main.zoom, Main.zoom);
    }

    private void drawRes (Graphics g, Res r){
        // for traffic simulation only
        if (r.load == 0)
            g.setColor (Color.red);
        else if (r.load == 1)
            g.setColor (Color.yellow);
        else
            g.setColor (Color.green);
        // g.setColor (Color.red);
        g.fillRect (r.x * Main.zoom, r.y * Main.zoom, Main.zoom, Main.zoom);
    }

    private void drawAgent (Graphics g, Agent a){
        g.setColor (Color.blue);
        g.fillRect (a.x * Main.zoom, a.y * Main.zoom, Main.zoom, Main.zoom);
    }

    protected void paintComponent (Graphics g){
        super.paintComponent (g);
        // draw walls
        for (Iterator e = files.walls.iterator () ; e.hasNext () ; )
            drawWall (g, (Wall)e.next ());
        // draw resources
        for (Iterator e = files.res.iterator () ; e.hasNext () ; )
            drawRes (g, (Res)e.next ());
        // draw agents (after resources, hence res may be hidden by agents)
        for (Iterator e = files.agents.iterator () ; e.hasNext () ; )
            drawAgent (g, (Agent)e.next ());
    }

    public String getToolTipText (MouseEvent m){
        // another solution would be to use a MouseListener which sets tooltip
        String tooltip;
        Res res = null;
        Agent ag = null;
        int y = m.getY () / Main.zoom;
        int x = m.getX () / Main.zoom;
        // find if there is a resource or an agent in square
        for (Iterator e = files.res.iterator () ; e.hasNext () ; ){
            Res r = (Res)e.next ();
            if ((r.y == y) && (r.x == x)){
                res = r;
                break;  // max 1 resource/square
            }
        }
        for (Iterator e = files.agents.iterator () ; e.hasNext () ; ){
            Agent a = (Agent)e.next ();
            if ((a.y == y) && (a.x == x)){
                ag = a;
                break;  // max 1 agent/square
            }
        }
        tooltip = "(" + y + "," + x + ")";
        if (res != null)  // found a resource
            tooltip += " R(" + res.id + "," + res.load + ")";
        if (ag != null)  // found an agent
            tooltip += " A(" + ag.id + "," + ag.load + ")";
        return tooltip;
    }
}

class IG extends JFrame
{
    MainPanel mainPanel;
    JTextField jtfCycle, jtfZoom;
    JButton goFwd, goBkw, goFastFwd, goFastBkw, goFirst, goLast;

    class GoLst implements ActionListener{
        public void actionPerformed (ActionEvent ae){
            try{
                int cycle = Integer.parseInt (jtfCycle.getText ());
                boolean isCycleReadFromFile = false;
                if (ae.getActionCommand().equals ("fwd"))
                    cycle ++;
                else if (ae.getActionCommand().equals ("bkw"))
                    cycle --;
                else if (ae.getActionCommand().equals ("first"))
                    cycle = 0;
                else if (ae.getActionCommand().equals ("last")){
                    while (! mainPanel.files.foundEOF){
                        cycle ++;
                        mainPanel.files.readAgentRes (cycle);
                    }
                    isCycleReadFromFile = true;
                }else if (ae.getActionCommand().equals ("fastbkw"))
                    while (--cycle > 0){
                        mainPanel.files.readAgentRes (cycle);
                        mainPanel.paintImmediately (mainPanel.getVisibleRect ());
                        jtfCycle.setText (Integer.toString (cycle));
                        jtfCycle.paintImmediately (jtfCycle.getVisibleRect ());
                    }
                else if (ae.getActionCommand().equals ("fastfwd")){
                    while (! mainPanel.files.foundEOF){
                        cycle ++;
                        mainPanel.files.readAgentRes (cycle);
                        if (! mainPanel.files.foundEOF){
                            mainPanel.paintImmediately (mainPanel.getVisibleRect ());
                            jtfCycle.setText (Integer.toString (cycle));
                            jtfCycle.paintImmediately (jtfCycle.getVisibleRect ());
                        }
                    }
                    isCycleReadFromFile = true;
                }  // else RET in cycle edit box => no need to change cycle
                if (! isCycleReadFromFile)
                    mainPanel.files.readAgentRes (cycle);
                mainPanel.repaint ();
                jtfCycle.setText (Integer.toString (cycle));
                enableButtons (cycle);
            }catch (NumberFormatException e){
                System.out.println ("ERROR: NumberFormatException - continue");
            }
        }
    }

    class ZoomLst implements ActionListener{
        public void actionPerformed (ActionEvent ae){
            try{
                Main.zoom = Integer.parseInt (jtfZoom.getText ());
                Main.mainWindow.pack();
            }catch (NumberFormatException e){
                System.out.println ("ERROR: NumberFormatException - continue");
                return;
            }
        }
    }

    class CtlPanel extends JPanel{
        CtlPanel (){
            setLayout (new GridLayout (7, 2));
            add (new JLabel ("BaseFileName:"));  // input file names
            add (new JLabel (Main.baseFileName));
            add (new JLabel ("Env. dim:"));  // env. dim
            add (new JLabel (Integer.toString (mainPanel.files.envy) + "x"
                             + Integer.toString (mainPanel.files.envx)));
            add (new JLabel ("Cycle:"));  // cycle
            jtfCycle = new JTextField (Integer.toString (Main.startCycle));
            GoLst go = new GoLst ();
            jtfCycle.addActionListener (go);
            // jtfCycle.setActionCommand ("cycle");
            add (jtfCycle);
            goBkw = new JButton ("<");  // bkw/fwd
            goBkw.addActionListener (go);
            goBkw.setActionCommand ("bkw");
            goBkw.setMnemonic ('b');
            goBkw.setToolTipText ("Bkw");
            add (goBkw);
            goFwd = new JButton (">");
            goFwd.addActionListener (go);
            goFwd.setActionCommand ("fwd");
            goFwd.setMnemonic ('f');
            goFwd.setToolTipText ("Fwd");
            add (goFwd);
            goFastBkw = new JButton ("<<");  // fast bkw/fwd
            goFastBkw.addActionListener (go);
            goFastBkw.setActionCommand ("fastbkw");
            goFastBkw.setToolTipText ("Fast bkw");
            add (goFastBkw);
            goFastFwd = new JButton (">>");
            goFastFwd.addActionListener (go);
            goFastFwd.setActionCommand ("fastfwd");
            goFastFwd.setToolTipText ("Fast fwd");
            add (goFastFwd);
            goFirst = new JButton ("|<");  // first/last
            goFirst.addActionListener (go);
            goFirst.setActionCommand ("first");
            goFirst.setToolTipText ("First cycle");
            add (goFirst);
            goLast = new JButton (">|");
            goLast.addActionListener (go);
            goLast.setActionCommand ("last");
            goLast.setToolTipText ("Last cycle");
            add (goLast);
            add (new JLabel ("Zoom:"));  // zoom
            jtfZoom = new JTextField (Integer.toString (Main.zoom));
            jtfZoom.addActionListener (new ZoomLst ());
            add (jtfZoom);
        }
    }

    private void enableButtons (int cycle){
        goBkw.setEnabled (cycle != 0);
        goFastBkw.setEnabled (cycle != 0);
        goFirst.setEnabled (cycle != 0);
        goFwd.setEnabled (! mainPanel.files.foundEOF);
        goFastFwd.setEnabled (! mainPanel.files.foundEOF);
        goLast.setEnabled (! mainPanel.files.foundEOF);
    }

    IG (){
        super ("ParSSAP visualiser");
        mainPanel = new MainPanel ();
        JPanel ctlPanel = new CtlPanel ();
        getContentPane().setLayout (new FlowLayout ());
        getContentPane().add (mainPanel);
        getContentPane().add (ctlPanel);
        enableButtons (Main.startCycle);
        setDefaultCloseOperation (EXIT_ON_CLOSE);
        pack ();
        goFwd.requestFocus ();
        setVisible (true);
    }
}

public class Main
{
    static int startCycle = 0;
    static int zoom = 2;
    static IG mainWindow;
    static String baseFileName = "Changes";
    static String envFileName, agFileName, resFileName;

    public static void main (String[] args) throws NumberFormatException{
        System.out.println
            ("Visualisation tool for ParSSAP output (pmSaveChanges()).");
        System.out.println ("Usage: prog [zoom [basefilename]]");
        if ((args.length == 1) || (args.length == 2))
            zoom = Integer.parseInt (args[0]);
        if (args.length == 2)
            baseFileName = args[1];
        envFileName = baseFileName + ".obs";
        agFileName = baseFileName + ".ags";
        resFileName = baseFileName + ".res";
        mainWindow = new IG ();
    }
}
